/*
 "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
 Copyright (C) 2001-2003 phonohawkput

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package aya;
import aya.*;
import java.util.*;
import java.util.regex.*;

public class LexicalAnalyzer {
    public static final int TYPE_IDENTIFIER = 0; // 識別子
    public static final int TYPE_SYMBOL = 1; // "if" "+" ">="等
    public static final int TYPE_NUMBER = 2; // 数値定数(符号無し)
    public static final int TYPE_STRING = 3; // 文字列定数(ダブルクオート無し)
    
    static Pattern space_of_head,comment_line,comment_block,identifier,symbol,const_string,const_number;
    static {
	// 先頭の空白。
	space_of_head = Pattern.compile("^\\s*");
	
	// 一行コメント。コメントは文字列定数よりも後に評価される事に注意。
	comment_line = Pattern.compile("^(?m://.*$)");
	
	// ブロックコメント。
	comment_block = Pattern.compile("^(?s:/\\*.*?\\*/)");
	
	// 識別子 (+ - * / = : ! ; { } % & # " ( ) [ ] < > , ? | ホワイトスペース を含まない文字列)
	// ただし予約語は除く。if,elseif,else,case,switch,when,others,while,for,break,continue,return
	// *** Java 1.4.2のjava.util.regexにはバグが！！ この\uffffを消すとマルチバイト文字が一切マッチしなくなる！！
	String ident_chars = "[^\uffff\\-+*/=:!;{}%&#\"()\\[\\]<>,?\\|\\s]";
	identifier = Pattern.compile("^"+ident_chars+"+");
	
	// シンボル名
	String symbol_plus_starting = "\\+\\+|\\+:=|\\+=|\\+";
	String symbol_minus_starting = "--|-:=|-=|-";
	String symbol_star_starting = "\\*:=|\\*=|\\*";
	String symbol_slash_starting = "/:=|/=|/";
	String symbol_equal_string = "==|=";
	String symbol_colon_starting = ":=|:";
	String symbol_excl_starting = "!_in_|!=";
	String symbol_parcent_starting = "%:=|%=|%";
	String symbol_lt_starting = "<=|<";
	String symbol_gt_starting = ">=|>";
	String symbol_reserved = "(?:if|elseif|else|case|switch|when|others|while|for|break|continue|return)(?!"+ident_chars+")"; // 識別子に使用可能な文字が後続しない。
	symbol = Pattern.compile("^(?:"+symbol_colon_starting+'|'+symbol_plus_starting+'|'+symbol_minus_starting+'|'+symbol_star_starting+'|'+symbol_slash_starting+'|'+symbol_equal_string+'|'+symbol_colon_starting+'|'+symbol_excl_starting+'|'+symbol_parcent_starting+'|'+symbol_lt_starting+'|'+symbol_gt_starting+'|'+symbol_reserved+"|_in_|&&|\\|\\||[&,;{}()\\[\\]])");
	
	// 文字列定数
	const_string = Pattern.compile("^\"(.*?)\"", Pattern.DOTALL);
	
	// 数値定数
	String hex_num = "0x[0-9a-fA-F]+";
	String bin_num = "0b[01]+";
	String dec_num = "[0-9]*\\.[0-9]+|[0-9]+";
	const_number = Pattern.compile("^(?:"+hex_num+"|"+bin_num+"|"+dec_num+")");
    }
    
    StringBuffer buf;
    Vector rollback_stack; // [[Token,Token,...][Token,Token,...]...]
    Vector rollbacked;
    boolean space_was_before_last_token;
    int lineNumber;
    
    public LexicalAnalyzer(String source) {
	buf = new StringBuffer(convertBackslashToYen(source));
	rollback_stack = new Vector();
	rollbacked = new Vector();
	space_was_before_last_token = false;
	lineNumber = 0;
	
	push();
    }
	
    public String toString() {
	String first = buf.substring(0,(buf.length() >= 100 ? 100 : buf.length()));
	return "<rollbacked: "+rollbacked+' '+"remaining(first 50): \""+first+"\">";
    }
	
    public void push() {
	rollback_stack.add(new Vector());
    }
    
    public void pop() {
	// スタックのトップを単に破棄する。
	rollback_stack.removeElementAt(rollback_stack.size()-1);
    }
    
    public void popnback() {
	// スタックのトップをロールバックして破棄する。
	// 後にロールバックされたトークンほど先に出る。
	rollbacked.addAll(0,(Vector)rollback_stack.elementAt(rollback_stack.size()-1));
	pop();
    }
    
    public void rollback() {
	// スタックのトップの最後のトークンをロールバックして破棄する。
	// 後にロールバックされたトークンほど先に出る。
	Vector top = (Vector)rollback_stack.elementAt(rollback_stack.size()-1);
	rollbacked.add(0,top.elementAt(top.size()-1));
	top.removeElementAt(top.size()-1);
    }
    
    public Token nextToken() throws Exception {
	// これ以上トークンが無かったらNoMoreTokensExceptionを投げる。
	
	// ロールバックされたトークンはあるか?
	Token result = null;
	if (rollbacked.size() > 0) {
	    result = (Token)rollbacked.elementAt(0);
	    rollbacked.removeElementAt(0);
	}
	else {
	    Matcher matcher;
	    while (true) {
		// 先頭のホワイトスペースを削除。
		if ((matcher = space_of_head.matcher(buf)).find()) {
		    space_was_before_last_token = (matcher.group().length() > 0);
		    // 改行の数を調べて行番号を増やす
		    for(int lastIndex = 0; lastIndex != -1; ) {
			lastIndex = matcher.group().indexOf('\n', lastIndex);
			if (lastIndex != -1) {
			    lineNumber++;
			    lastIndex++;
			}
		    }
		    buf.delete(0,matcher.group().length());
		}
		
		// 空になったか?
		if (buf.length() == 0) {
		    throw new NoMoreTokensException();
		}
		/*if (buf.toString().startsWith("\u305d\u306e\u4ed6")) {
			System.out.println("CHECK IT!");
		}*/
		
		if ((matcher = const_string.matcher(buf)).find()) {
		    // まずは文字列定数を評価。ここで確定。
		    result = new Token(TYPE_STRING,matcher.group(1),space_was_before_last_token,lineNumber);
		    buf.delete(0,matcher.group().length());
		    break;
		}
		else if ((matcher = comment_line.matcher(buf)).find()) {
		    // 次にコメントを評価。
		    buf.delete(0,matcher.group().length());
		}
		else if ((matcher = comment_block.matcher(buf)).find()) {
		    buf.delete(0,matcher.group().length());
		}
		else if ((matcher = symbol.matcher(buf)).find()) {
		    // シンボルを評価。
		    result = new Token(TYPE_SYMBOL,matcher.group(),space_was_before_last_token,lineNumber);
		    buf.delete(0,matcher.group().length());
		    break;
		}
		else if ((matcher = identifier.matcher(buf)).find()) {
		    // 識別子を評価。
		    String token = matcher.group();
		    Matcher submatcher;
		    if ((submatcher = const_number.matcher(token)).find() &&
			submatcher.group().length() == token.length()) {
			// 識別子のやうに見えたが、それを構成する文字全てが
			// 數値定數としての形を滿してゐれば、それは數値定數である。
			result = new Token(TYPE_NUMBER,token,space_was_before_last_token,lineNumber);
			// 數値定數よりも先に識別子を評價してゐるのは、
			// 數字で始まる名前を持つ識別子が、なんと文では
			// 正當なものである爲。
		    }
		    else {
			result = new Token(TYPE_IDENTIFIER,token,space_was_before_last_token,lineNumber);
		    }
		    buf.delete(0,token.length());
		    break;
		}
		else if ((matcher = const_number.matcher(buf)).find()) {
		    // 数値定数を評価。
		    result = new Token(TYPE_NUMBER,matcher.group(),space_was_before_last_token,lineNumber);
		    buf.delete(0,matcher.group().length());
		    break;
		}
		else {
		    int size = buf.length() >= 15 ? 15 : buf.length();
		    throw new Exception(
			"Syntax error occured at line "+lineNumber+", near \""+buf.substring(0,size)+"\"");
		}
	    }
	}
	
	if (result != null) {
	    // スタックのトップのバッファに登録
	    Vector top = (Vector)rollback_stack.elementAt(rollback_stack.size()-1);
	    top.add(result);
	}
	return result;
    }
    
    public static String convertBackslashToYen(String src) {
	// バックスラッシュを￥に置き換える
	StringBuffer buf = new StringBuffer();
	final int len = src.length();
	for (int i = 0;i < len;i++) {
	    char c = src.charAt(i);
	    int n = (int)c;
	    buf.append(c == 0x005c ? 0x00a5 : c);
	}
	return buf.toString();
    }
}
